import sys

from typing import Any, Dict, Optional

from clipped.config.exceptions import ClippedException, SchemaError

PolyaxonException = ClippedException
PolyaxonSchemaError = SchemaError


class OpenApiException(PolyaxonException):
    """The base exception class for all OpenAPIExceptions"""


class ApiTypeError(OpenApiException, TypeError):
    def __init__(self, msg, path_to_item=None, valid_classes=None, key_type=None):
        """Raises an exception for TypeErrors

        Args:
            msg (str): the exception message

        Keyword Args:
            path_to_item (list): a list of keys an indices to get to the
                                 current_item
                                 None if unset
            valid_classes (tuple): the primitive classes that current item
                                   should be an instance of
                                   None if unset
            key_type (bool): False if our value is a value in a dict
                             True if it is a key in a dict
                             False if our item is an item in a list
                             None if unset
        """
        self.path_to_item = path_to_item
        self.valid_classes = valid_classes
        self.key_type = key_type
        full_msg = msg
        if path_to_item:
            full_msg = "{0} at {1}".format(msg, render_path(path_to_item))
        super(ApiTypeError, self).__init__(full_msg)


class ApiValueError(OpenApiException, ValueError):
    def __init__(self, msg, path_to_item=None):
        """
        Args:
            msg (str): the exception message

        Keyword Args:
            path_to_item (list) the path to the exception in the
                received_data dict. None if unset
        """

        self.path_to_item = path_to_item
        full_msg = msg
        if path_to_item:
            full_msg = "{0} at {1}".format(msg, render_path(path_to_item))
        super(ApiValueError, self).__init__(full_msg)


class ApiKeyError(OpenApiException, KeyError):
    def __init__(self, msg, path_to_item=None):
        """
        Args:
            msg (str): the exception message

        Keyword Args:
            path_to_item (None/list) the path to the exception in the
                received_data dict
        """
        self.path_to_item = path_to_item
        full_msg = msg
        if path_to_item:
            full_msg = "{0} at {1}".format(msg, render_path(path_to_item))
        super(ApiKeyError, self).__init__(full_msg)


class ApiException(OpenApiException):
    def __init__(self, status=None, reason=None, http_resp=None):
        if http_resp:
            self.status = http_resp.status
            self.reason = http_resp.reason
            self.body = http_resp.data
            self.headers = http_resp.getheaders()
        else:
            self.status = status
            self.reason = reason
            self.body = None
            self.headers = None

    def __str__(self):
        """Custom error messages for exception"""
        error_message = "({0})\n" "Reason: {1}\n".format(self.status, self.reason)
        if self.headers:
            error_message += "HTTP response headers: {0}\n".format(self.headers)

        if self.body:
            error_message += "HTTP response body: {0}\n".format(self.body)

        return error_message


class NotFoundException(ApiException):
    def __init__(self, status=None, reason=None, http_resp=None):
        super(NotFoundException, self).__init__(status, reason, http_resp)


class UnauthorizedException(ApiException):
    def __init__(self, status=None, reason=None, http_resp=None):
        super(UnauthorizedException, self).__init__(status, reason, http_resp)


class ForbiddenException(ApiException):
    def __init__(self, status=None, reason=None, http_resp=None):
        super(ForbiddenException, self).__init__(status, reason, http_resp)


class ServiceException(ApiException):
    def __init__(self, status=None, reason=None, http_resp=None):
        super(ServiceException, self).__init__(status, reason, http_resp)


def render_path(path_to_item):
    """Returns a string representation of a path"""
    result = ""
    for pth in path_to_item:
        if isinstance(pth, int):
            result += "[{0}]".format(pth)
        else:
            result += "['{0}']".format(pth)
    return result


def handle_api_error(
    e,
    logger: Any,
    message: Optional[str] = None,
    http_messages_mapping: Optional[Dict] = None,
    sys_exit: bool = False,
):
    if http_messages_mapping:
        http_messages_mapping.update(HTTP_ERROR_MESSAGES_MAPPING)
    else:
        http_messages_mapping = HTTP_ERROR_MESSAGES_MAPPING
    if message:
        logger.error(message)
    if e and hasattr(e, "status"):
        if e.status not in http_messages_mapping.keys():
            logger.error("Exception:")
            logger.error(e, stack_info=True, exc_info=True)
        elif getattr(e, "body") and e.status != 404:
            logger.error("Error: %s" % e.body)
        if getattr(e, "reason"):
            logger.error("Reason: %s" % e.reason)
        message = http_messages_mapping.get(e.status)
        if message:
            logger.error(message)
    elif e and hasattr(e, "message"):  # Handling of HTML errors
        error_found = False
        for k in http_messages_mapping.keys():
            if str(k) in e.message:
                logger.error(http_messages_mapping.get(k))
                error_found = True
                break
        if not error_found:
            logger.error("Error:")
            logger.error(e.message)
    elif e:
        logger.error("Exception:")
        logger.error(e, stack_info=True, exc_info=True)
    if sys_exit:
        sys.exit(1)


class PolyaxonOperatorException(PolyaxonException):
    def __init__(self, cmd, args, return_code, stdout, stderr):
        self.cmd = cmd
        self.args = args
        self.return_code = return_code
        self.stdout = stdout.read() if stdout else None
        self.stderr = stderr.read()
        if stdout:
            message = "`{}` command {} failed with exit status {}\nstdout:\n{}\nstderr:\n{}".format(
                self.cmd, self.args, self.return_code, self.stdout, self.stderr
            )
        else:
            message = "`{}` command {} failed with exit status {}\nstderr:\n{}".format(
                self.cmd, self.args, self.return_code, self.stderr
            )
        super().__init__(message=message)


class PolyaxonConverterError(PolyaxonException):
    pass


class PolyaxonCompilerError(PolyaxonException):
    pass


class PolyTuneException(PolyaxonException):
    pass


class PolyaxonConfigException(PolyaxonException):
    pass


class PolyaxonK8sError(PolyaxonException):
    pass


class PolyaxonAgentError(PolyaxonException):
    pass


class PolyaxonBuildException(PolyaxonException):
    pass


class PolyaxonContainerException(Exception):
    pass


class PolyaxonConnectionError(PolyaxonException):
    pass


class PolyaxonPathException(PolyaxonException):
    pass


class PQLException(PolyaxonException):
    pass


class PolyaxonValidationError(PolyaxonSchemaError):
    pass


class PolyaxonfileError(PolyaxonSchemaError):
    pass


class PolyaxonClientException(OpenApiException):
    """The base exception class for all client exceptions"""


class PolyaxonShouldExitError(PolyaxonClientException):
    pass


class PolyaxonHTTPError(PolyaxonClientException):
    def __init__(self, endpoint, response, message=None, status_code=None):
        super().__init__()
        self.endpoint = endpoint
        self.response = response
        self.message = getattr(self, "message", message)
        self.status_code = getattr(self, "status_code", status_code)

    def __str__(self):
        return "{status_code} on {endpoint}.".format(
            status_code=self.status_code, endpoint=self.endpoint
        )


HTTP_ERROR_MESSAGES_MAPPING = {
    400: "Status: 400. One or more request parameters are incorrect",
    401: "Status: 401. Authentication failed. Retry by invoking Polyaxon login.",
    403: "Status: 403. You are not authorized to access this resource on Polyaxon.",
    404: "Status: 404. "
    "The resource you are looking for was not found. Check if the name or uuid is correct.",
    405: "Status: 405. Endpoint does not exist or not configured on this API, "
    "make sure you are connecting to the correct host.",
    429: "Status: 429. You are over the allowed limits for this operation.",
    500: "Status: 502. Internal server error, please try again later.",
    502: "Status: 502. Invalid response from Polyaxon server.",
    503: "Status: 503. A problem was encountered, please try again later.",
    504: "Status: 504. Polyaxon server took too long to respond.",
    525: "Status: 525. SSL error.",
}


def handle_api_error(
    e,
    logger: Any,
    message: Optional[str] = None,
    http_messages_mapping: Optional[Dict] = None,
    sys_exit: bool = False,
):
    if http_messages_mapping:
        http_messages_mapping.update(HTTP_ERROR_MESSAGES_MAPPING)
    else:
        http_messages_mapping = HTTP_ERROR_MESSAGES_MAPPING
    if message:
        logger.error(message)
    if e and hasattr(e, "status"):
        if e.status not in http_messages_mapping.keys():
            logger.error("Exception:")
            logger.error(e, stack_info=True, exc_info=True)
        elif getattr(e, "body") and e.status != 404:
            logger.error("Error: %s" % e.body)
        if getattr(e, "reason"):
            logger.error("Reason: %s" % e.reason)
        message = http_messages_mapping.get(e.status)
        if message:
            logger.error(message)
    elif e and hasattr(e, "message"):  # Handling of HTML errors
        error_found = False
        for k in http_messages_mapping.keys():
            if str(k) in e.message:
                logger.error(http_messages_mapping.get(k))
                error_found = True
                break
        if not error_found:
            logger.error("Error:")
            logger.error(e.message)
    elif e:
        logger.error("Exception:")
        logger.error(e, stack_info=True, exc_info=True)
    if sys_exit:
        sys.exit(1)



from typing import List, Optional

from clipped.compact.pydantic import StrictStr

from polyaxon._schemas.base import BaseSchemaModel


class V1ArtifactsMount(BaseSchemaModel):
    _IDENTIFIER = "artifacts_mount"

    name: StrictStr
    paths: Optional[List[StrictStr]]



from typing import Optional

from clipped.compact.pydantic import StrictStr
from clipped.config.schema import BaseAllowSchemaModel
from clipped.types.uuids import UUIDStr

from polyaxon._flow import V1CloningKind


class V1Cloning(BaseAllowSchemaModel):
    uuid: Optional[UUIDStr]
    name: Optional[StrictStr]
    kind: Optional[V1CloningKind]



from datetime import datetime
from typing import Any, Dict, List, Optional

from polyaxon import settings
from polyaxon._compiler.contexts import resolve_contexts, resolve_globals_contexts
from polyaxon._compiler.lineage import resolve_artifacts_lineage
from polyaxon._compiler.resolver import AgentResolver
from polyaxon._flow import V1CloningKind, V1CompiledOperation, V1Operation, V1RunKind
from polyaxon._polyaxonfile import CompiledOperationSpecification
from polyaxon.exceptions import PolyaxonCompilerError


class BaseResolver:
    KINDS = {
        V1RunKind.JOB,
        V1RunKind.SERVICE,
        V1RunKind.MPIJOB,
        V1RunKind.TFJOB,
        V1RunKind.PADDLEJOB,
        V1RunKind.PYTORCHJOB,
        V1RunKind.MXJOB,
        V1RunKind.XGBJOB,
        V1RunKind.NOTIFIER,
        V1RunKind.CLEANER,
        V1RunKind.TUNER,
        V1RunKind.WATCHDOG,
        V1RunKind.DAG,
        V1RunKind.DASKJOB,
        V1RunKind.RAYJOB,
        V1RunKind.TUNER,
    }

    def __init__(
        self,
        run,
        compiled_operation: V1CompiledOperation,
        owner_name: str,
        project_name: str,
        project_uuid: str,
        run_name: str,
        run_uuid: str,
        run_path: str,
        params: Optional[Dict],
        created_at: Optional[datetime] = None,
        compiled_at: Optional[datetime] = None,
        cloning_kind: V1CloningKind = None,
        original_uuid: Optional[str] = None,
        is_independent: bool = True,
    ):
        if not compiled_operation:
            raise PolyaxonCompilerError("A run spec is required for resolution.")
        self.run = run
        self.compiled_operation = compiled_operation
        self.owner_name = owner_name
        self.project_name = project_name
        self.project_uuid = project_uuid
        self.project_uuid = project_uuid or project_name
        self.run_name = run_name
        self.run_uuid = run_uuid or run_name
        self.run_path = run_path
        self.params = params or {}
        self.connection_by_names = {}
        self.namespace = None
        self.artifacts_store = None
        self.secrets = None
        self.config_maps = None
        self.polyaxon_sidecar = None
        self.polyaxon_init = None
        self.iteration = None
        self.agent_config = None
        self.globals = {}
        self.artifacts = []
        self.created_at = created_at
        self.compiled_at = compiled_at
        self.schedule_at = None
        self.started_at = None
        self.finished_at = None
        self.duration = None
        self.cloning_kind = cloning_kind
        self.original_uuid = original_uuid
        self.is_independent = is_independent
        self._param_spec = {}

    @classmethod
    def is_valid(cls, compiled_operation: V1CompiledOperation):
        compiled_operation.validate_build()
        run_kind = compiled_operation.get_run_kind()
        if run_kind not in cls.KINDS:
            raise PolyaxonCompilerError(
                "Resolver Error. "
                "Specification with run kind: {} "
                "is not supported in this deployment version.".format(run_kind)
            )

    @property
    def param_spec(self):
        return self._param_spec

    def resolve_edges(self):
        pass

    def resolve_globals_contexts(self):
        self.globals = resolve_globals_contexts(
            namespace=self.namespace,
            owner_name=self.owner_name,
            project_name=self.project_name,
            project_uuid=self.project_uuid,
            run_uuid=self.run_uuid,
            run_name=self.run_name,
            run_path=self.run_path,
            iteration=self.iteration,
            created_at=self.created_at,
            compiled_at=self.compiled_at,
            schedule_at=self.schedule_at,
            started_at=self.started_at,
            finished_at=self.finished_at,
            duration=self.duration,
            cloning_kind=self.cloning_kind,
            original_uuid=self.original_uuid,
            is_independent=self.is_independent,
        )

    def resolve_params(self):
        pass

    def apply_params(self, should_be_resolved: bool = True):
        self.compiled_operation = CompiledOperationSpecification.apply_params(
            config=self.compiled_operation,
            params=self.params,
            context=self.globals,
        )
        self._param_spec = CompiledOperationSpecification.calculate_context_spec(
            config=self.compiled_operation,
            contexts=self.globals,
            should_be_resolved=should_be_resolved,
        )

    def resolve_connections_params(self):
        self.compiled_operation = (
            CompiledOperationSpecification.apply_run_connections_params(
                config=self.compiled_operation,
                artifact_store=self.agent_config.artifacts_store.name
                if self.agent_config
                else None,
                contexts=self.globals,
                param_spec=self.param_spec,
            )
        )

    def resolve_presets(self):
        pass

    def resolve_agent(self):
        if settings.AGENT_CONFIG:
            self.agent_config = settings.AGENT_CONFIG.clone()

    def apply_operation_contexts(self):
        try:
            self.compiled_operation = (
                CompiledOperationSpecification.apply_operation_contexts(
                    self.compiled_operation,
                    param_spec=self.param_spec,
                    contexts=self.globals,
                )
            )
        except Exception as e:
            raise PolyaxonCompilerError(
                "Could not apply run context, error: {}".format(repr(e))
            )

    def resolve_io(self):
        pass

    def resolve_access(self):
        pass

    def resolve_connections(self):
        agent_env = AgentResolver.construct()
        agent_env.resolve(
            compiled_operation=self.compiled_operation, agent_config=self.agent_config
        )
        self.polyaxon_sidecar = agent_env.polyaxon_sidecar
        self.polyaxon_init = agent_env.polyaxon_init
        self.namespace = agent_env.namespace
        self.secrets = agent_env.secrets
        self.config_maps = agent_env.config_maps
        self.connection_by_names = agent_env.connection_by_names
        self.artifacts_store = agent_env.artifacts_store

    def resolve_init_refs(self):
        pass

    def clean_init_refs(self):
        pass

    def resolve_actions(self):
        pass

    def resolve_artifacts_lineage(self):
        self.artifacts += resolve_artifacts_lineage(
            owner_name=self.owner_name,
            project_name=self.project_name,
            project_uuid=self.project_uuid,
            run_name=self.run_name,
            run_path=self.run_path,
            run_uuid=self.run_uuid,
            param_spec=self.param_spec,
            compiled_operation=self.compiled_operation,
            connection_by_names=self.connection_by_names,
            artifacts_store=self.artifacts_store,
        )

    def _resolve_contexts(self):
        return resolve_contexts(
            namespace=self.namespace,
            owner_name=self.owner_name,
            project_name=self.project_name,
            project_uuid=self.project_uuid,
            run_uuid=self.run_uuid,
            run_name=self.run_name,
            run_path=self.run_path,
            compiled_operation=self.compiled_operation,
            artifacts_store=self.artifacts_store,
            connection_by_names=self.connection_by_names,
            iteration=self.iteration,
            created_at=self.created_at,
            compiled_at=self.compiled_at,
            schedule_at=self.schedule_at,
            started_at=self.started_at,
            finished_at=self.finished_at,
            duration=self.duration,
            cloning_kind=self.cloning_kind,
            original_uuid=self.original_uuid,
            is_independent=self.is_independent,
        )

    def _apply_runtime_contexts(self):
        contexts = self._resolve_contexts()
        return CompiledOperationSpecification.apply_runtime_contexts(
            self.compiled_operation, contexts=contexts
        )

    def _apply_pipeline_contexts(self):
        return self.compiled_operation

    def _should_skip_runtime_resolution(self):
        return False

    def apply_runtime_contexts(self):
        if self._should_skip_runtime_resolution():
            return
        if self.compiled_operation.has_pipeline:
            self.compiled_operation = self._apply_pipeline_contexts()
        else:
            self.compiled_operation = self._apply_runtime_contexts()

    def resolve_state(self):
        pass

    def persist_state(self):
        pass

    def resolve_build(self):
        raise PolyaxonCompilerError(
            "Build resolution is not supported in this Polyaxon distribution"
        )

    def resolve_hooks(self) -> List[V1Operation]:
        raise PolyaxonCompilerError(
            "Hooks resolution is not supported in this Polyaxon distribution"
        )

    def resolve(self) -> V1CompiledOperation:
        self.resolve_edges()
        self.resolve_globals_contexts()
        self.resolve_params()
        self.apply_params()
        self.resolve_presets()
        self.resolve_agent()
        self.resolve_connections_params()
        self.apply_operation_contexts()
        self.resolve_io()
        self.resolve_init_refs()
        self.resolve_access()
        self.resolve_connections()
        self.resolve_artifacts_lineage()
        self.clean_init_refs()
        self.resolve_state()
        self.apply_runtime_contexts()
        self.persist_state()
        return self.compiled_operation



import os

from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

from clipped.utils.http import clean_host
from clipped.utils.lists import to_list
from vents.connections.connection_schema import patch_git

from polyaxon import settings
from polyaxon._auxiliaries import V1PolyaxonInitContainer, V1PolyaxonSidecarContainer
from polyaxon._connections import (
    CONNECTION_CONFIG,
    V1Connection,
    V1ConnectionKind,
    V1ConnectionResource,
)
from polyaxon._containers.names import INIT_PREFIX, SIDECAR_PREFIX
from polyaxon._contexts import paths as ctx_paths
from polyaxon._docker import docker_types
from polyaxon._env_vars.keys import (
    ENV_KEYS_ARTIFACTS_STORE_NAME,
    ENV_KEYS_COLLECT_ARTIFACTS,
    ENV_KEYS_COLLECT_RESOURCES,
    ENV_KEYS_LOG_LEVEL,
    ENV_KEYS_NO_API,
    ENV_KEYS_RUN_INSTANCE,
)
from polyaxon._flow import V1CompiledOperation, V1Init, V1Plugins
from polyaxon._k8s import k8s_schemas
from polyaxon._runner.converter.common import constants
from polyaxon._runner.converter.common.containers import ensure_container_name
from polyaxon._runner.converter.common.volumes import get_volume_name
from polyaxon._runner.converter.types import (
    Container,
    ContainerPort,
    EnvVar,
    Resource,
    ResourceRequirements,
    VolumeMount,
)
from polyaxon._runner.kinds import RunnerKind
from polyaxon._schemas.types import (
    V1ArtifactsType,
    V1DockerfileType,
    V1FileType,
    V1TensorboardType,
)
from polyaxon._services.auth import AuthenticationTypes
from polyaxon._services.headers import PolyaxonServiceHeaders
from polyaxon._services.values import PolyaxonServices
from polyaxon._utils.fqn_utils import get_resource_name, get_run_instance
from polyaxon._utils.host_utils import get_api_host
from polyaxon.exceptions import PolyaxonConverterError, PolyaxonSchemaError


class BaseConverter:
    RUNNER_KIND: RunnerKind = None
    SPEC_KIND: Optional[str] = None
    MAIN_CONTAINER_ID: Optional[str] = None

    def __init__(
        self,
        owner_name: str,
        project_name: str,
        run_name: str,
        run_uuid: str,
        run_path: Optional[str] = None,
        namespace: str = "default",
        internal_auth: bool = False,
        polyaxon_sidecar: V1PolyaxonSidecarContainer = None,
        polyaxon_init: V1PolyaxonInitContainer = None,
        base_env_vars: bool = False,
    ):
        self.is_valid()
        self.owner_name = owner_name
        self.project_name = project_name
        self.run_name = run_name
        self.run_uuid = run_uuid
        self.run_path = run_path or self.run_uuid
        self.resource_name = self.get_resource_name()
        self.run_instance = self.get_instance()
        self.namespace = namespace
        self.internal_auth = internal_auth
        self.base_env_vars = base_env_vars
        self.polyaxon_sidecar = polyaxon_sidecar
        self.polyaxon_init = polyaxon_init

    def get_instance(self) -> str:
        return get_run_instance(
            owner=self.owner_name, project=self.project_name, run_uuid=self.run_uuid
        )

    def get_resource_name(self) -> str:
        return get_resource_name(self.run_uuid)

    def is_valid(self):
        if not self.RUNNER_KIND:
            raise PolyaxonConverterError(
                "Please make sure that a converter subclass has a valid RUNNER_KIND"
            )
        if not self.SPEC_KIND:
            raise PolyaxonConverterError(
                "Please make sure that a converter subclass has a valid SPEC_KIND"
            )
        if not self.MAIN_CONTAINER_ID:
            raise PolyaxonConverterError(
                "Please make sure that a converter subclass has a valid MAIN_CONTAINER_ID"
            )

    @staticmethod
    def get_by_name(values: List[Any]) -> Dict[str, Any]:
        return {c.name: c for c in values}

    @staticmethod
    def _post_process_host(host: str) -> str:
        return host

    @classmethod
    def get_api_host(cls, external_host: bool = False) -> str:
        if external_host:
            return cls._post_process_host(
                get_api_host(default=settings.CLIENT_CONFIG.host)
            )
        else:
            return cls._post_process_host(clean_host(settings.CLIENT_CONFIG.host))

    @staticmethod
    def filter_connections_from_init(init: List[V1Init]) -> List[V1Init]:
        return [i for i in init if i.has_connection()]

    @staticmethod
    def filter_containers_from_init(init: List[V1Init]) -> List[Container]:
        return [i.container for i in init if not i.has_connection()]

    @staticmethod
    def _sanitize_container_env(
        env: List[EnvVar],
    ) -> Optional[List[EnvVar]]:
        raise NotImplementedError

    @classmethod
    def _sanitize_container(cls, container: Container) -> Container:
        raise NotImplementedError

    @classmethod
    def _patch_container(
        cls,
        container: Container,
        name: Optional[str] = None,
        command: Optional[List[str]] = None,
        args: Optional[List[str]] = None,
        image: Optional[str] = None,
        env: Optional[List[EnvVar]] = None,
        volume_mounts: Optional[List[VolumeMount]] = None,
        ports: Optional[List[ContainerPort]] = None,
        resources: Optional[ResourceRequirements] = None,
        **kwargs,
    ) -> k8s_schemas.V1Container:
        raise NotImplementedError

    def _get_service_env_vars(
        self,
        service_header: str,
        header: Optional[str] = None,
        include_secret_key: bool = False,
        include_internal_token: bool = False,
        include_agent_token: bool = False,
        authentication_type: Optional[str] = None,
        external_host: bool = False,
        log_level: Optional[str] = None,
    ) -> List[EnvVar]:
        raise NotImplementedError

    def _get_polyaxon_sidecar_service_env_vars(
        self,
        external_host: bool = False,
        log_level: Optional[str] = None,
    ) -> List[EnvVar]:
        if not self.base_env_vars:
            return self._get_service_env_vars(
                service_header=PolyaxonServices.SIDECAR,
                authentication_type=AuthenticationTypes.TOKEN,
                header=PolyaxonServiceHeaders.SERVICE,
                external_host=external_host,
                log_level=log_level,
            )
        env = []
        if settings.CLIENT_CONFIG.no_api:
            env += [self._get_env_var(name=ENV_KEYS_NO_API, value=True)]
        if log_level:
            env += [self._get_env_var(name=ENV_KEYS_LOG_LEVEL, value=log_level)]
        proxy_env = self._get_proxy_env_vars(
            settings.AGENT_CONFIG.use_proxy_env_vars_use_in_ops
        )
        if proxy_env:
            env += proxy_env
        return env

    def _get_auth_service_env_vars(
        self,
        external_host: bool = False,
        log_level: Optional[str] = None,
    ) -> Optional[List[EnvVar]]:
        if self.base_env_vars:
            return None
        return self._get_service_env_vars(
            service_header=PolyaxonServices.INITIALIZER,
            include_internal_token=self.internal_auth,
            include_agent_token=not self.internal_auth,
            authentication_type=(
                AuthenticationTypes.INTERNAL_TOKEN
                if self.internal_auth
                else AuthenticationTypes.TOKEN
            ),
            header=(
                PolyaxonServiceHeaders.INTERNAL
                if self.internal_auth
                else PolyaxonServiceHeaders.SERVICE
            ),
            external_host=external_host,
            log_level=log_level,
        )

    def _get_init_service_env_vars(
        self,
        external_host: bool = False,
        log_level: Optional[str] = None,
    ) -> Optional[List[EnvVar]]:
        if not self.base_env_vars:
            return self._get_service_env_vars(
                service_header=PolyaxonServices.INITIALIZER,
                authentication_type=AuthenticationTypes.TOKEN,
                header=PolyaxonServiceHeaders.SERVICE,
                external_host=external_host,
                log_level=log_level,
            )
        env = []
        if settings.CLIENT_CONFIG.no_api:
            env.append(self._get_env_var(name=ENV_KEYS_NO_API, value=True))
        if log_level:
            env.append(self._get_env_var(name=ENV_KEYS_LOG_LEVEL, value=log_level))
        proxy_env = self._get_proxy_env_vars(
            settings.AGENT_CONFIG.use_proxy_env_vars_use_in_ops
        )
        if proxy_env:
            env += proxy_env
        return env

    @classmethod
    def _get_run_instance_env_var(cls, run_instance: str) -> EnvVar:
        return cls._get_env_var(name=ENV_KEYS_RUN_INSTANCE, value=run_instance)

    @staticmethod
    def _get_env_var(name: str, value: Any) -> EnvVar:
        raise NotImplementedError

    @classmethod
    def _get_connections_catalog_env_var(
        cls,
        connections: List[V1Connection],
    ) -> Optional[EnvVar]:
        catalog = CONNECTION_CONFIG.get_connections_catalog(connections)
        if not catalog:
            return None
        return cls._get_env_var(
            name=CONNECTION_CONFIG.get_connections_catalog_env_name(),
            value=catalog.to_json(),
        )

    @staticmethod
    def _get_connection_env_var(connection: V1Connection) -> List[EnvVar]:
        env_vars = []
        if not connection:
            return env_vars

        if connection.env:
            env_vars += to_list(connection.env, check_none=True)

        return env_vars

    @staticmethod
    def _get_proxy_env_var(key: str) -> Optional[str]:
        value = os.environ.get(key)
        if not value:
            value = os.environ.get(key.lower())
        if not value:
            value = os.environ.get(key.upper())

        return value

    @classmethod
    def _add_proxy_env_var(cls, name: str, value: str) -> List[EnvVar]:
        return [
            cls._get_env_var(name.upper(), value),
            cls._get_env_var(name, value),
        ]

    @classmethod
    def _get_proxy_env_vars(
        cls,
        use_proxy_env_vars_use_in_ops: bool,
    ) -> List[docker_types.V1EnvVar]:
        if use_proxy_env_vars_use_in_ops:
            env_vars = []
            https_proxy = cls._get_proxy_env_var("HTTPS_PROXY")
            if not https_proxy:
                https_proxy = cls._get_proxy_env_var("https_proxy")
            if https_proxy:
                env_vars += cls._add_proxy_env_var(
                    name="HTTPS_PROXY", value=https_proxy
                )
                env_vars += cls._add_proxy_env_var(
                    name="https_proxy", value=https_proxy
                )
            http_proxy = cls._get_proxy_env_var("HTTP_PROXY")
            if not http_proxy:
                http_proxy = cls._get_proxy_env_var("http_proxy")
            if http_proxy:
                env_vars += cls._add_proxy_env_var(name="HTTP_PROXY", value=http_proxy)
                env_vars += cls._add_proxy_env_var(name="http_proxy", value=http_proxy)
            no_proxy = cls._get_proxy_env_var("NO_PROXY")
            if not no_proxy:
                no_proxy = cls._get_proxy_env_var("no_proxy")
            if no_proxy:
                env_vars += cls._add_proxy_env_var(name="NO_PROXY", value=no_proxy)
                env_vars += cls._add_proxy_env_var(name="no_proxy", value=no_proxy)
            return env_vars
        return []

    @classmethod
    def _get_kv_env_vars(cls, kv_env_vars: List[List]) -> List[EnvVar]:
        env_vars = []
        if not kv_env_vars:
            return env_vars

        for kv_env_var in kv_env_vars:
            if not kv_env_var or not len(kv_env_var) == 2:
                raise PolyaxonConverterError(
                    "Received a wrong a key value env var `{}`".format(kv_env_var)
                )
            env_vars.append(cls._get_env_var(name=kv_env_var[0], value=kv_env_var[1]))

        return env_vars

    @classmethod
    def _get_env_vars_from_k8s_resources(
        cls,
        secrets: Iterable[V1ConnectionResource],
        config_maps: Iterable[V1ConnectionResource],
    ) -> List[EnvVar]:
        raise NotImplementedError

    @classmethod
    def _get_additional_env_vars(cls) -> List[EnvVar]:
        raise NotImplementedError

    @staticmethod
    def _get_base_env_vars(
        namespace: str,
        resource_name: str,
        use_proxy_env_vars_use_in_ops: bool,
        log_level: Optional[str],
    ) -> List[EnvVar]:
        raise NotImplementedError

    @staticmethod
    def _new_container(name: str) -> Container:
        raise NotImplementedError

    @classmethod
    def _get_sidecar_container(
        cls,
        container_id: str,
        polyaxon_sidecar: V1PolyaxonSidecarContainer,
        env: List[EnvVar],
        artifacts_store: V1Connection,
        plugins: V1Plugins,
        run_path: Optional[str],
    ) -> Optional[Container]:
        raise NotImplementedError

    @classmethod
    def _ensure_container(
        cls, container: Container, volumes: List[k8s_schemas.V1Volume]
    ) -> docker_types.V1Container:
        raise NotImplementedError

    def _get_main_env_vars(
        self,
        plugins: V1Plugins,
        kv_env_vars: List[List],
        artifacts_store_name: str,
        connections: Iterable[V1Connection],
        secrets: Iterable[V1ConnectionResource],
        config_maps: Iterable[V1ConnectionResource],
    ) -> List[EnvVar]:
        if self.base_env_vars:
            env = self._get_base_env_vars(
                namespace=self.namespace,
                resource_name=self.resource_name,
                use_proxy_env_vars_use_in_ops=settings.AGENT_CONFIG.use_proxy_env_vars_use_in_ops,
                log_level=plugins.log_level if plugins else None,
            )
        else:
            env = self._get_service_env_vars(
                service_header=PolyaxonServices.RUNNER,
                external_host=plugins.external_host if plugins else False,
                log_level=plugins.log_level if plugins else None,
            )
        env = to_list(env, check_none=True)
        connections = connections or []

        if plugins and plugins.collect_artifacts:
            env.append(self._get_env_var(name=ENV_KEYS_COLLECT_ARTIFACTS, value=True))

        if plugins and plugins.collect_resources:
            env.append(self._get_env_var(name=ENV_KEYS_COLLECT_RESOURCES, value=True))

        if artifacts_store_name:
            env.append(
                self._get_env_var(
                    name=ENV_KEYS_ARTIFACTS_STORE_NAME, value=artifacts_store_name
                )
            )

        # Add connections catalog env vars information
        env += to_list(
            self._get_connections_catalog_env_var(connections=connections),
            check_none=True,
        )
        # Add connection env vars information
        for connection in connections:
            try:
                env += to_list(
                    self._get_connection_env_var(connection=connection), check_none=True
                )
            except PolyaxonSchemaError as e:
                raise PolyaxonConverterError("Error resolving secrets: %s" % e) from e

        env += self._get_kv_env_vars(kv_env_vars)
        env += self._get_env_vars_from_k8s_resources(
            secrets=secrets, config_maps=config_maps
        )
        env += self._get_additional_env_vars()
        return env

    @classmethod
    def _get_docker_context_mount(cls) -> VolumeMount:
        raise NotImplementedError

    @staticmethod
    def _get_auth_context_mount(
        read_only: Optional[bool] = None,
        run_path: Optional[str] = None,
    ) -> VolumeMount:
        raise NotImplementedError

    @staticmethod
    def _get_artifacts_context_mount(
        read_only: Optional[bool] = None,
        run_path: Optional[str] = None,
    ) -> VolumeMount:
        raise NotImplementedError

    @staticmethod
    def _get_connections_context_mount(
        name: str, mount_path: str, run_path: str
    ) -> VolumeMount:
        raise NotImplementedError

    @staticmethod
    def _get_shm_context_mount() -> VolumeMount:
        raise NotImplementedError

    @classmethod
    def _get_mount_from_store(cls, store: V1Connection) -> Optional[VolumeMount]:
        raise NotImplementedError

    @staticmethod
    def _get_mount_from_resource(
        resource: V1ConnectionResource,
    ) -> Optional[VolumeMount]:
        raise NotImplementedError

    @staticmethod
    def _get_volume(host_path: str, mount_path: str, read_only: bool) -> VolumeMount:
        raise NotImplementedError

    @classmethod
    def _get_mounts(
        cls,
        use_auth_context: bool,
        use_docker_context: bool,
        use_shm_context: bool,
        use_artifacts_context: bool,
        run_path: Optional[str] = None,
    ) -> List[VolumeMount]:
        raise NotImplementedError

    @classmethod
    def _get_main_volume_mounts(
        cls,
        run_path: str,
        plugins: V1Plugins,
        init: Optional[List[V1Init]],
        connections: Iterable[V1Connection],
        secrets: Iterable[V1ConnectionResource],
        config_maps: Iterable[V1ConnectionResource],
    ) -> List[VolumeMount]:
        init = init or []
        connections = connections or []
        secrets = secrets or []
        config_maps = config_maps or []

        volume_mounts = []
        volume_names = set()
        if plugins and plugins.collect_artifacts:
            volume_mounts += to_list(
                cls._get_artifacts_context_mount(read_only=False, run_path=run_path),
                check_none=True,
            )
            volume_names.add(constants.VOLUME_MOUNT_ARTIFACTS)
        for init_connection in init:
            volume_name = (
                get_volume_name(init_connection.path)
                if init_connection.path
                else constants.VOLUME_MOUNT_ARTIFACTS
            )
            mount_path = init_connection.path or ctx_paths.CONTEXT_MOUNT_ARTIFACTS
            if volume_name in volume_names:
                continue
            volume_names.add(volume_name)
            volume_mounts += to_list(
                cls._get_connections_context_mount(
                    name=volume_name, mount_path=mount_path, run_path=run_path
                ),
                check_none=True,
            )
        for store in connections:
            volume_mounts += to_list(
                cls._get_mount_from_store(store=store), check_none=True
            )

        for secret in secrets:
            volume_mounts += to_list(
                cls._get_mount_from_resource(resource=secret), check_none=True
            )

        for config_map in config_maps:
            volume_mounts += to_list(
                cls._get_mount_from_resource(resource=config_map), check_none=True
            )

        return volume_mounts

    def _get_main_container(
        self,
        container_id: str,
        main_container: Container,
        plugins: V1Plugins,
        artifacts_store: Optional[V1Connection],
        init: Optional[List[V1Init]],
        connections: Optional[List[str]],
        connection_by_names: Dict[str, V1Connection],
        secrets: Optional[Iterable[V1ConnectionResource]],
        config_maps: Optional[Iterable[V1ConnectionResource]],
        run_path: Optional[str],
        kv_env_vars: List[List] = None,
        ports: List[int] = None,
    ) -> Container:
        raise NotImplementedError

    @classmethod
    def _get_base_store_container(
        cls,
        container: Optional[docker_types.V1Container],
        container_name: str,
        polyaxon_init: V1PolyaxonInitContainer,
        store: V1Connection,
        env: List[docker_types.V1EnvVar],
        env_from: List[k8s_schemas.V1EnvFromSource],
        volume_mounts: List[docker_types.V1VolumeMount],
        args: List[str],
        command: Optional[List[str]] = None,
    ) -> Optional[docker_types.V1Container]:
        raise NotImplementedError

    @classmethod
    def _get_custom_init_container(
        cls,
        connection: V1Connection,
        plugins: V1Plugins,
        container: Optional[Container],
        run_path: str,
        env: List[EnvVar] = None,
        mount_path: Optional[str] = None,
    ) -> Container:
        raise NotImplementedError

    @classmethod
    def _get_dockerfile_init_container(
        cls,
        polyaxon_init: V1PolyaxonInitContainer,
        dockerfile_args: V1DockerfileType,
        plugins: V1Plugins,
        run_path: str,
        run_instance: str,
        container: Optional[Container] = None,
        env: List[EnvVar] = None,
        mount_path: Optional[str] = None,
    ) -> Container:
        raise NotImplementedError

    @classmethod
    def _get_file_init_container(
        cls,
        polyaxon_init: V1PolyaxonInitContainer,
        file_args: V1FileType,
        plugins: V1Plugins,
        run_path: str,
        run_instance: str,
        container: Optional[Container] = None,
        env: List[EnvVar] = None,
        mount_path: Optional[str] = None,
    ) -> Container:
        raise NotImplementedError

    @classmethod
    def _get_git_init_container(
        cls,
        polyaxon_init: V1PolyaxonInitContainer,
        connection: V1Connection,
        plugins: V1Plugins,
        run_path: str,
        container: Optional[Container] = None,
        env: List[EnvVar] = None,
        mount_path: Optional[str] = None,
        track: bool = False,
    ) -> Container:
        raise NotImplementedError

    @classmethod
    def _get_store_init_container(
        cls,
        polyaxon_init: V1PolyaxonInitContainer,
        connection: V1Connection,
        artifacts: V1ArtifactsType,
        paths: Union[List[str], List[Tuple[str, str]]],
        run_path: str,
        container: Optional[Container] = None,
        env: List[EnvVar] = None,
        mount_path: Optional[str] = None,
        is_default_artifacts_store: bool = False,
    ) -> Container:
        raise NotImplementedError

    @classmethod
    def _get_tensorboard_init_container(
        cls,
        polyaxon_init: V1PolyaxonInitContainer,
        artifacts_store: V1Connection,
        tb_args: V1TensorboardType,
        plugins: V1Plugins,
        run_path: str,
        run_instance: str,
        container: Optional[Container] = None,
        env: List[EnvVar] = None,
        mount_path: Optional[str] = None,
    ) -> Container:
        raise NotImplementedError

    @classmethod
    def _get_auth_context_init_container(
        cls,
        polyaxon_init: V1PolyaxonInitContainer,
        run_path: str,
        env: Optional[List[EnvVar]] = None,
    ) -> Container:
        raise NotImplementedError

    @classmethod
    def _get_artifacts_path_init_container(
        cls,
        polyaxon_init: V1PolyaxonInitContainer,
        artifacts_store: V1Connection,
        run_path: str,
        auto_resume: bool,
        env: Optional[List[EnvVar]] = None,
    ) -> Container:
        raise NotImplementedError

    def _handle_init_connections(
        self,
        polyaxon_init: V1PolyaxonInitContainer,
        artifacts_store: V1Connection,
        init_connections: List[V1Init],
        connection_by_names: Dict[str, V1Connection],
        plugins: V1Plugins,
        log_level: Optional[str] = None,
        volumes: Optional[List[k8s_schemas.V1Volume]] = None,
    ) -> List[Container]:
        containers = []
        external_host = plugins.external_host if plugins else False

        # Prepare connections that Polyaxon can init automatically
        for init_connection in init_connections:
            if init_connection.connection:
                connection_spec = connection_by_names.get(init_connection.connection)
                # Handling ssh with git
                if (
                    V1ConnectionKind.is_ssh(connection_spec.kind)
                    and init_connection.git
                ):
                    patch_git(connection_spec.schema_, init_connection.git)
                    containers.append(
                        self._get_git_init_container(
                            run_path=self.run_path,
                            polyaxon_init=polyaxon_init,
                            connection=connection_spec,
                            container=self._ensure_container(
                                init_connection.container, volumes=volumes
                            ),
                            env=self._get_init_service_env_vars(
                                external_host=external_host,
                                log_level=log_level,
                            ),
                            mount_path=init_connection.path,
                            plugins=plugins,
                            track=True,
                        )
                    )
                elif V1ConnectionKind.is_git(connection_spec.kind):
                    if init_connection.git:  # Update the default schema
                        connection_spec.schema_.patch(init_connection.git)
                    containers.append(
                        self._get_git_init_container(
                            run_path=self.run_path,
                            polyaxon_init=polyaxon_init,
                            connection=connection_spec,
                            container=self._ensure_container(
                                init_connection.container, volumes=volumes
                            ),
                            env=self._get_init_service_env_vars(
                                external_host=external_host,
                                log_level=log_level,
                            ),
                            mount_path=init_connection.path,
                            plugins=plugins,
                            track=True,
                        )
                    )
                elif V1ConnectionKind.is_artifact(connection_spec.kind):
                    containers.append(
                        self._get_store_init_container(
                            run_path=self.run_path,
                            polyaxon_init=polyaxon_init,
                            connection=connection_spec,
                            artifacts=init_connection.artifacts,
                            paths=init_connection.paths,
                            container=self._ensure_container(
                                init_connection.container, volumes=volumes
                            ),
                            env=self._get_init_service_env_vars(
                                external_host=external_host,
                                log_level=log_level,
                            ),
                            mount_path=init_connection.path,
                            is_default_artifacts_store=artifacts_store
                            and init_connection.connection == artifacts_store.name,
                        )
                    )
                else:
                    containers.append(
                        self._get_custom_init_container(
                            run_path=self.run_path,
                            connection=connection_spec,
                            container=self._ensure_container(
                                init_connection.container, volumes=volumes
                            ),
                            env=self._get_init_service_env_vars(
                                external_host=external_host,
                                log_level=log_level,
                            ),
                            mount_path=init_connection.path,
                            plugins=plugins,
                        )
                    )
            else:
                # artifacts init without connection should default to the artifactsStore
                if init_connection.artifacts or init_connection.paths:
                    containers.append(
                        self._get_store_init_container(
                            run_path=self.run_path,
                            polyaxon_init=polyaxon_init,
                            connection=artifacts_store,
                            artifacts=init_connection.artifacts,
                            paths=init_connection.paths,
                            container=self._ensure_container(
                                init_connection.container, volumes=volumes
                            ),
                            env=self._get_init_service_env_vars(
                                external_host=external_host,
                                log_level=log_level,
                            ),
                            mount_path=init_connection.path,
                            is_default_artifacts_store=True,
                        )
                    )
                # git init without connection
                if init_connection.git:
                    git_name = init_connection.git.get_name()
                    containers.append(
                        self._get_git_init_container(
                            run_path=self.run_path,
                            polyaxon_init=polyaxon_init,
                            connection=V1Connection(
                                name=git_name,
                                kind=V1ConnectionKind.GIT,
                                schema_=init_connection.git,
                                secret=None,
                            ),
                            container=self._ensure_container(
                                init_connection.container, volumes=volumes
                            ),
                            env=self._get_init_service_env_vars(
                                external_host=external_host,
                                log_level=log_level,
                            ),
                            mount_path=init_connection.path,
                            plugins=plugins,
                            track=False,
                        )
                    )
                # Dockerfile initialization
                if init_connection.dockerfile:
                    containers.append(
                        self._get_dockerfile_init_container(
                            polyaxon_init=polyaxon_init,
                            dockerfile_args=init_connection.dockerfile,
                            env=self._get_init_service_env_vars(
                                external_host=external_host,
                                log_level=log_level,
                            ),
                            mount_path=init_connection.path,
                            container=self._ensure_container(
                                init_connection.container, volumes=volumes
                            ),
                            plugins=plugins,
                            run_path=self.run_path,
                            run_instance=self.run_instance,
                        )
                    )
                # File initialization
                if init_connection.file:
                    containers.append(
                        self._get_file_init_container(
                            polyaxon_init=polyaxon_init,
                            file_args=init_connection.file,
                            env=self._get_init_service_env_vars(
                                external_host=external_host,
                                log_level=log_level,
                            ),
                            mount_path=init_connection.path,
                            container=self._ensure_container(
                                init_connection.container, volumes=volumes
                            ),
                            plugins=plugins,
                            run_path=self.run_path,
                            run_instance=self.run_instance,
                        )
                    )
                # Tensorboard initialization
                if init_connection.tensorboard:
                    containers.append(
                        self._get_tensorboard_init_container(
                            polyaxon_init=polyaxon_init,
                            artifacts_store=artifacts_store,
                            tb_args=init_connection.tensorboard,
                            env=self._get_init_service_env_vars(
                                external_host=external_host,
                                log_level=log_level,
                            ),
                            mount_path=init_connection.path,
                            container=self._ensure_container(
                                init_connection.container, volumes=volumes
                            ),
                            plugins=plugins,
                            run_path=self.run_path,
                            run_instance=self.run_instance,
                        )
                    )

        return containers

    def get_init_containers(
        self,
        polyaxon_init: V1PolyaxonInitContainer,
        plugins: V1Plugins,
        artifacts_store: V1Connection,
        init_connections: List[V1Init],
        init_containers: List[Container],
        connection_by_names: Dict[str, V1Connection],
        log_level: Optional[str] = None,
        volumes: Optional[List[k8s_schemas.V1Volume]] = None,
    ) -> List[Container]:
        init_containers = [
            self._ensure_container(
                ensure_container_name(container=c, prefix=INIT_PREFIX), volumes=volumes
            )
            for c in to_list(init_containers, check_none=True)
        ]
        init_connections = to_list(init_connections, check_none=True)
        containers = []

        # Add auth context
        if plugins and plugins.auth:
            containers.append(
                self._get_auth_context_init_container(
                    polyaxon_init=polyaxon_init,
                    run_path=self.run_path,
                    env=self._get_auth_service_env_vars(
                        external_host=plugins.external_host
                    ),
                )
            )

        # Add outputs
        if plugins and plugins.collect_artifacts:
            containers += to_list(
                self._get_artifacts_path_init_container(
                    polyaxon_init=polyaxon_init,
                    artifacts_store=artifacts_store,
                    run_path=self.run_path,
                    auto_resume=plugins.auto_resume,
                    env=self._get_proxy_env_vars(
                        settings.AGENT_CONFIG.use_proxy_env_vars_use_in_ops
                    ),
                ),
                check_none=True,
            )

        containers += self._handle_init_connections(
            polyaxon_init=polyaxon_init,
            artifacts_store=artifacts_store,
            init_connections=init_connections,
            connection_by_names=connection_by_names,
            plugins=plugins,
            log_level=log_level,
            volumes=volumes,
        )
        return containers + init_containers

    def get_sidecar_containers(
        self,
        polyaxon_sidecar: V1PolyaxonSidecarContainer,
        plugins: V1Plugins,
        artifacts_store: V1Connection,
        sidecar_containers: List[Container],
        log_level: Optional[str] = None,
        volumes: Optional[List[k8s_schemas.V1Volume]] = None,
    ) -> List[Container]:
        sidecar_containers = [
            ensure_container_name(container=c, prefix=SIDECAR_PREFIX)
            for c in to_list(sidecar_containers, check_none=True)
        ]
        polyaxon_sidecar_container = self._get_sidecar_container(
            container_id=self.MAIN_CONTAINER_ID,
            polyaxon_sidecar=polyaxon_sidecar,
            env=self._get_polyaxon_sidecar_service_env_vars(
                external_host=plugins.external_host if plugins else False,
                log_level=log_level,
            ),
            artifacts_store=artifacts_store,
            plugins=plugins,
            run_path=self.run_path,
        )
        containers = to_list(polyaxon_sidecar_container, check_none=True)
        containers += sidecar_containers
        return [self._ensure_container(c, volumes) for c in containers]

    def get_main_container(
        self,
        main_container: Container,
        plugins: V1Plugins,
        artifacts_store: V1Connection,
        connections: List[str],
        init_connections: Optional[List[V1Init]],
        connection_by_names: Dict[str, V1Connection],
        secrets: Optional[Iterable[V1ConnectionResource]],
        config_maps: Optional[Iterable[V1ConnectionResource]],
        kv_env_vars: Optional[List[List]] = None,
        ports: Optional[List[int]] = None,
    ) -> Container:
        return self._get_main_container(
            container_id=self.MAIN_CONTAINER_ID,
            main_container=main_container,
            plugins=plugins,
            artifacts_store=artifacts_store,
            connections=connections,
            init=init_connections,
            connection_by_names=connection_by_names,
            secrets=secrets,
            config_maps=config_maps,
            kv_env_vars=kv_env_vars,
            ports=ports,
            run_path=self.run_path,
        )

    def get_resource(
        self,
        compiled_operation: V1CompiledOperation,
        artifacts_store: V1Connection,
        connection_by_names: Dict[str, V1Connection],
        secrets: Optional[Iterable[V1ConnectionResource]],
        config_maps: Optional[Iterable[V1ConnectionResource]],
        default_sa: Optional[str] = None,
        default_auth: bool = False,
    ) -> Resource:
        raise NotImplementedError



from typing import List, Optional

from polyaxon._contexts import paths as ctx_paths
from polyaxon._schemas.types import V1FileType

FILE_INIT_COMMAND = ["polyaxon", "initializer", "file"]


def get_file_init_args(
    file_args: V1FileType,
    run_path: str,
    mount_path: Optional[str] = None,
) -> List[str]:
    return [
        "--file-context={}".format(file_args.to_json()),
        "--filepath={}".format(mount_path),
        "--copy-path={}".format(
            ctx_paths.CONTEXT_MOUNT_RUN_OUTPUTS_FORMAT.format(run_path)
        ),
        "--track",
    ]



import os

from datetime import datetime
from typing import Dict, List, Optional, Set, Tuple

from clipped.utils.dates import path_last_modified
from clipped.utils.paths import get_files_and_dirs_in_path

from polyaxon._contexts import paths as ctx_paths
from polyaxon._schemas.base import BaseSchemaModel


class PathData(BaseSchemaModel):
    __root__: Tuple[str, datetime, str]

    class Config(BaseSchemaModel.Config):
        validate_assignment = False

    @property
    def base(self) -> str:
        return self.__root__[0]

    @property
    def ts(self) -> datetime:
        if isinstance(self.__root__[1], str):
            self.__root__ = (
                self.__root__[0],
                datetime.fromisoformat(self.__root__[1]),
                self.__root__[1],
            )
        return self.__root__[1]

    @property
    def op(self) -> str:
        return self.__root__[2]


class FSWatcher(BaseSchemaModel):
    _IDENTIFIER = "fswatcher"

    _PUT = "put"
    _RM = "rm"
    _NOOP = ""

    dir_mapping: Optional[Dict[str, PathData]]
    file_mapping: Optional[Dict[str, PathData]]

    @property
    def dirs_mp(self) -> Dict[str, PathData]:
        return self.dir_mapping or {}

    @property
    def files_mp(self) -> Dict[str, PathData]:
        return self.file_mapping or {}

    class Config(BaseSchemaModel.Config):
        validate_assignment = False

    @staticmethod
    def delete(path: str):
        if os.path.exists(path):
            os.remove(path)

    def write(self, filepath: str, mode: Optional[int] = None):
        filepath = filepath or ctx_paths.CONTEXT_MOUNT_FILE_WATCHER
        return super().write(filepath=filepath, mode=mode)

    def _sync_path(self, path: str, base_path: str, mapping: Dict) -> Dict:
        current_ts = path_last_modified(path)
        rel_path = os.path.relpath(path, base_path)
        data = mapping.get(rel_path)
        if data:
            if current_ts > data.ts:
                mapping[rel_path] = PathData.construct(
                    __root__=(base_path, current_ts, self._PUT)
                )
            else:
                mapping[rel_path] = PathData.construct(
                    __root__=(base_path, data.ts, self._NOOP)
                )
        else:
            mapping[rel_path] = PathData.construct(
                __root__=(base_path, current_ts, self._PUT)
            )
        return mapping

    def sync_file(self, path: str, base_path: str):
        self.file_mapping = self._sync_path(path, base_path, self.files_mp)

    def sync_dir(self, path: str, base_path: str):
        self.dir_mapping = self._sync_path(path, base_path, self.dirs_mp)

    def init(self):
        self.dir_mapping = {
            p: PathData.construct(__root__=(d.base, d.ts, self._RM))
            for p, d in self.dirs_mp.items()
        }
        self.file_mapping = {
            p: PathData.construct(__root__=(d.base, d.ts, self._RM))
            for p, d in self.files_mp.items()
        }

    def sync(self, path: str, exclude: Optional[List[str]] = None):
        files, dirs = get_files_and_dirs_in_path(
            path, exclude=exclude, collect_dirs=True
        )
        base_path, prefix_path = os.path.split(path)
        for file_path in files:
            self.sync_file(os.path.join(prefix_path, file_path), base_path=base_path)

        for dir_path in dirs:
            self.sync_dir(os.path.join(prefix_path, dir_path), base_path=base_path)

    def _get_mapping_by_op(self, mapping: Dict, op: str) -> Set:
        return {(p.base, k) for k, p in mapping.items() if p.op == op}

    def _clean_by_op(self, mapping: Dict, op: str) -> Dict:
        return {k: p for k, p in mapping.items() if p.op != op}

    def get_files_to_put(self) -> Set:
        return self._get_mapping_by_op(self.files_mp, self._PUT)

    def get_files_to_rm(self) -> Set:
        results = self._get_mapping_by_op(self.files_mp, self._RM)
        self.file_mapping = self._clean_by_op(self.files_mp, self._RM)
        return results

    def get_dirs_to_put(self) -> Set:
        return self._get_mapping_by_op(self.dirs_mp, self._PUT)

    def get_dirs_to_rm(self) -> Set:
        results = self._get_mapping_by_op(self.dirs_mp, self._RM)
        self.dir_mapping = self._clean_by_op(self.dirs_mp, self._RM)
        return results



from typing import Dict, Iterable, List, Optional

from polyaxon._connections import V1Connection, V1ConnectionResource
from polyaxon._docker import docker_types
from polyaxon._docker.converter.base import BaseConverter
from polyaxon._docker.converter.mixins import ServiceMixin
from polyaxon._flow import V1CompiledOperation, V1Plugins, V1Service


class ServiceConverter(ServiceMixin, BaseConverter):
    def get_resource(
        self,
        compiled_operation: V1CompiledOperation,
        artifacts_store: V1Connection,
        connection_by_names: Dict[str, V1Connection],
        secrets: Optional[Iterable[V1ConnectionResource]],
        config_maps: Optional[Iterable[V1ConnectionResource]],
        default_sa: Optional[str] = None,
        default_auth: bool = False,
    ) -> List[docker_types.V1Container]:
        service = compiled_operation.run  # type: V1Service
        plugins = V1Plugins.get_or_create(
            config=compiled_operation.plugins, auth=default_auth
        )
        kv_env_vars = compiled_operation.get_env_io()
        return self.get_replica_resource(
            plugins=plugins,
            environment=service.environment,
            volumes=service.volumes,
            init=service.init,
            sidecars=service.sidecars,
            container=service.container,
            artifacts_store=artifacts_store,
            connections=service.connections,
            connection_by_names=connection_by_names,
            secrets=secrets,
            config_maps=config_maps,
            kv_env_vars=kv_env_vars,
            default_sa=default_sa,
            ports=service.ports,
        )
